
# $Id: dpkt.py 43 2007-08-02 22:42:59Z jon.oberheide $

"""Simple packet creation and parsing."""

import array, itertools, socket, struct

class Error(Exception): pass
class UnpackError(Error): pass
class NeedData(UnpackError): pass
class PackError(Error): pass

class MetaPacket(type):
    def __init__(cls, name, bases, dct):
        st = getattr(cls, '__hdr__', None)
        if st is not None:
            cls.__hdr_fields__ = [x[0] for x in st]
            cls.__hdr_struct__ = struct.Struct(
                    getattr(cls, '__byte_order__', '>') + \
                    ''.join(x[1] for x in st))
            cls.__hdr_fmt__ = cls.__hdr_struct__.format
            cls.__hdr_len__ = cls.__hdr_struct__.size
            cls.__hdr_defaults__ = dict(itertools.izip(cls.__hdr_fields__, (x[2] for x in st)))

class Packet(object):
    """Base packet class, with metaclass magic to generate members from
    self.__hdr__.

    __hdr__ should be defined as a list of (name, structfmt, default) tuples
    __byte_order__ can be set to override the default ('>')

    Example::

    >>> class Foo(Packet):
    ...   __hdr__ = (('foo', 'I', 1), ('bar', 'H', 2), ('baz', '4s', 'quux'))
    ...
    >>> foo = Foo(bar=3)
    >>> foo
    Foo(bar=3)
    >>> str(foo)
    '\x00\x00\x00\x01\x00\x03quux'
    >>> foo.bar
    3
    >>> foo.baz
    'quux'
    >>> foo.foo = 7
    >>> foo.baz = 'whee'
    >>> foo
    Foo(baz='whee', foo=7, bar=3)
    >>> Foo('hello, world!')
    Foo(baz=' wor', foo=1751477356L, bar=28460, data='ld!')
    """
    __metaclass__ = MetaPacket

    def __init__(self, *args, **kwargs):
        """Packet constructor with ([buf], [field=val,...]) prototype.

        Arguments:

        buf -- optional packet buffer to unpack

        Optional keyword arguments correspond to members to set
        (matching fields in self.__hdr__, or 'data').
        """
        if args:
            try:
                self.unpack(args[0])
            except struct.error:
                if len(args[0]) < self.__hdr_len__:
                    raise NeedData
                raise UnpackError('invalid %s: %r' %
                                  (self.__class__.__name__, args[0]))
        else:
            self.data = ''
            dct = self.__dict__
            dct.update(self.__hdr_defaults__)
            dct.update(kwargs)

    def __len__(self):
        return self.__hdr_len__ + len(self.data)

    def __repr__(self):
        dct = self.__dict__
        l = ['%s=%r' % (k, dct[k])
             for k in self.__hdr_fields__
             if dct[k] != self.__hdr_defaults__[k]]
        if self.data:
            l.append('data=%r' % self.data)
        return '%s(%s)' % (self.__class__.__name__, ', '.join(l))

    def __str__(self):
        return self.pack_hdr() + str(self.data)

    def pack_hdr(self):
        """Return packed header string."""
        dct = self.__dict__
        try:
            return self.__hdr_struct__.pack(*[dct[k] for k in self.__hdr_fields__])
        except struct.error, e:
            raise PackError(str(e))

    def pack(self):
        """Return packed header + self.data string."""
        return str(self)

    def unpack(self, buf):
        """Unpack packet header fields from buf, and set self.data."""
        dct = self.__dict__
        for k, v in itertools.izip(self.__hdr_fields__,
                self.__hdr_struct__.unpack(buf[:self.__hdr_len__])):
            dct[k] = v
        self.data = buf[self.__hdr_len__:]

def in_cksum(buf):
    """Return computed Internet checksum."""
    if len(buf) % 2 == 1:
        buf += '\x00'
    a = array.array('H', buf)
    s = sum(a)
    s = (s >> 16) + (s & 0xffff)
    s += (s >> 16)
    return socket.ntohs(~s & 0xffff)

class MetaAddress(type):
    def __init__(cls, name, bases, dct):
        super(MetaAddress, cls).__init__(name, bases, dct)
        cls.cache = {}

class Address(str):
    __metaclass__ = MetaAddress
    def __new__(cls, n=None, ascii=None):
        cache = cls.cache
        if n in cache:
            return cache[n]
        if ascii is not None:
            n = cls.aton(ascii)
            i = cls.ntoi(n)
        elif isinstance(n, (int, long)):
            i, n = n, cls.iton(n)
            ascii = cls.ntoa(n)
        else:
            i = cls.ntoi(n)
            ascii = cls.ntoa(n)
        addr = super(Address, cls).__new__(cls, n)
        addr.i = i
        addr.ascii = ascii
        cache[i] = addr
        cache[n] = addr
        return addr
    def __int__(self):
        return self.i
    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.ascii)
