"""Easily generate sequences of packets."""

__author__ = 'Jeffrey Knockel'
__copyright__ = 'Copyright 2013 Jeffrey Knockel'
__license__ = 'GPL v2'

from contextlib import closing
import os
import random
import socket
import time

from dpkt.ip import IP, IPAddress
from dpkt.ip6 import IP6, IP6Address, IP6FragmentHeader
from dpkt.icmp import ICMP, ICMP_ECHO
from dpkt.icmp6 import ICMP6, ICMP6_ECHO_REQUEST
from dpkt.tcp import TCP, TH_ACK, TH_RST, TH_SYN, TCPFlags

PID = os.getpid()
SO_BINDTODEVICE = 25

def default_src4(dst, iface=None):
    """Returns the default IPv4 source address for a given destination.

    If the optional 'iface' is given, the source address will be from that
    interface.
    """
    with closing(socket.socket(socket.AF_INET, socket.SOCK_RAW, socket.IPPROTO_RAW)) as s:
        if iface is not None:
            s.setsockopt(socket.SOL_SOCKET, SO_BINDTODEVICE, iface + '\x00')
        s.connect((dst.ascii, 0))
        src = s.getsockname()[0]
    return IPAddress(ascii=src)

def default_src6(dst, iface=None):
    """Returns the default IPv6 source address for a given destination.

    If the optional 'iface' is given, the source address will be from that
    interface.
    """
    with closing(socket.socket(socket.AF_INET6, socket.SOCK_RAW, socket.IPPROTO_RAW)) as s:
        if iface is not None:
            s.setsockopt(socket.SOL_SOCKET, SO_BINDTODEVICE, iface + '\x00')
        s.connect((dst.ascii, 0))
        src = s.getsockname()[0]
    return IP6Address(ascii=src)

def _fill_data(seq, length):
    if length > 0:
        packed = '%.4x%.4x' % (PID, seq.data)
        seq.data = seq.data + 1 & 0xffff
        return packed * (length // 8) + packed[:length % 8]
    return ''

class MetaSequence(type):
    def __init__(cls, name, bases, dct):
        super(MetaSequence, cls).__init__(name, bases, dct)
        cls.cache = {}
    def __call__(cls, dst, src=None, iface=None, *args, **kwargs):
        if src is None:
            src = cls.default_src(dst, iface)
        pair = (dst, src)
        if pair in cls.cache:
            return cls.cache[pair]
        new = cls.__new__(cls, dst, src, *args, **kwargs)
        new.__init__(dst, src, *args, **kwargs)
        return cls.cache.setdefault(pair, new)

class Sequence(object):
    __metaclass__ = MetaSequence

class IPSequence(Sequence):
    """Generates sequences of IPv4 packets."""
    def __init__(self, ip_dst, ip_src, ip_id=None):
        if ip_id is None:
            ip_id = int(time.time())
        self.ip_dst = ip_dst
        self.ip_src = ip_src
        self.ip_id = ip_id
        self.sanitize_ip_id()
    default_src = staticmethod(default_src4)
    def next_ip_id(self):
        ip_id = self.ip_id
        self.ip_id += 1
        self.sanitize_ip_id()
        return ip_id
    def sanitize_ip_id(self):
        self.ip_id &= 0xffff
        if self.ip_id == 0:
            # with an AF_INET*/SOCK_RAW socket, the linux kernel will fill in
            # zero'd ip id's with its own
            self.ip_id = 1
    def next_packet_testing_mtu(self, mtu, **kwargs):
        return self.next_packet(mtu - 20, **kwargs)

class IP6Sequence(Sequence):
    """Generates sequences of IPv6 packets."""
    def __init__(self, ip_dst, ip_src, ip_id=None):
        if ip_id is None:
            ip_id = int(time.time())
        self.ip_dst = ip_dst
        self.ip_src = ip_src
        self.ip_id = ip_id
        self.sanitize_ip_id()
    default_src = staticmethod(default_src6)
    def next_ip_id(self):
        ip_id = self.ip_id
        self.ip_id += 1
        self.sanitize_ip_id()
        return ip_id
    def sanitize_ip_id(self):
        self.ip_id &= 0xffffffff
        if self.ip_id == 0:
            # with an AF_INET*/SOCK_RAW socket, the linux kernel will fill in
            # zero'd ip id's with its own
            self.ip_id = 1
    def next_packet(self, *args, **kwargs):
        raise NotImplementedError()
    def next_packet_testing_mtu(self, mtu, **kwargs):
        # mtu - 40, not mtu - 48, since we want to measure when the other host
        # begins including the fragmentation header in its response, despite
        # the fact that we always include it
        return self.next_packet(mtu - 40, **kwargs)

class ICMPSequence(IPSequence):
    """Generates sequences of ICMP packets."""
    def __init__(self, ip_dst, ip_src=None, ip_id=None, icmp_id=PID, icmp_seq=1, data=1):
        super(ICMPSequence, self).__init__(ip_dst, ip_src, ip_id)
        self.icmp_id = icmp_id
        self.icmp_seq = icmp_seq & 0xffff
        self.data = data & 0xffff
    def next_packet(self, data_length=64, type=ICMP_ECHO):
        data_length = max(data_length, 0)
        p = IP(len=data_length + 8, p=socket.IPPROTO_ICMP, id=self.next_ip_id(), dst=self.ip_dst, src=self.ip_src, data=
               ICMP(type=type, data=
               ICMP.Echo(id=self.icmp_id, seq=self.icmp_seq, data=
               _fill_data(self, data_length))))
        self.icmp_seq = self.icmp_seq + 1 & 0xffff
        return p
    def next_packet_testing_mtu(self, mtu, *args, **kwargs):
        return super(ICMPSequence, self).next_packet_testing_mtu(mtu - 8, *args, **kwargs)

class ICMP6Sequence(IP6Sequence):
    """Generates sequences of ICMPv6 packets."""
    def __init__(self, ip_dst, ip_src=None, ip_id=None, icmp_id=PID, icmp_seq=1, data=1):
        super(ICMP6Sequence, self).__init__(ip_dst, ip_src, ip_id)
        self.icmp_id = icmp_id
        self.icmp_seq = icmp_seq & 0xffff
        self.data = data & 0xffff
    def next_packet(self, data_length=64, type=ICMP6_ECHO_REQUEST):
        data_length = max(data_length, 0)
        p = IP6(plen=data_length + 16, nxt=44, dst=self.ip_dst, src=self.ip_src, p=socket.IPPROTO_ICMPV6, data=
                ICMP6(type=type, data=
                ICMP6.Echo(id=self.icmp_id, seq=self.icmp_seq, data=
                _fill_data(self, data_length))))
        p.extension_hdrs[44] = IP6FragmentHeader(nxt=socket.IPPROTO_ICMPV6, id=self.next_ip_id())
        self.icmp_seq = self.icmp_seq + 1 & 0xffff
        return p
    def next_packet_testing_mtu(self, mtu, *args, **kwargs):
        return super(ICMP6Sequence, self).next_packet_testing_mtu(mtu - 8, *args, **kwargs)

class TCPSequenceMixin(object):
    def rst(self, p):
        p.id = self.next_ip_id()
        p.data.flags = TCPFlags(TH_RST)
        p.data.win = 0
        p.data.seq = (p.data.seq + 1) & 0xffffffff
        p.data.ack = 0
        p.data.sum = 0

class TCPSequence(IPSequence, TCPSequenceMixin):
    """Generates sequences of TCP/IPv4 packets."""
    def __init__(self, ip_dst, ip_src=None, ip_id=None, tcp_sport=None, tcp_dport=None, tcp_seq_random=None, tcp_ack_random=None, data=1):
        super(TCPSequence, self).__init__(ip_dst, ip_src, ip_id)
        if tcp_sport is None:
            tcp_sport = int(time.time())
        self.tcp_sport = (tcp_sport - 1024) % (32768 - 1024) + 1024
        if tcp_dport is None:
            tcp_dport = int(time.time()) + (32768 - 1024) // 2
        self.tcp_dport = (tcp_dport - 1024) % (32768 - 1024) + 1024
        if tcp_seq_random is None:
            tcp_seq_random = random.Random()
        self.tcp_seq_random = tcp_seq_random
        if tcp_ack_random is None:
            tcp_ack_random = random.Random()
        self.tcp_ack_random = tcp_ack_random
        self.data = data & 0xffff
    def next_packet(self, data_length=0, sport=None, dport=None, flags=TCPFlags(TH_SYN)):
        data_length = max(data_length, 0)
        if sport is None:
            sport = self.tcp_sport
            self.tcp_sport = (self.tcp_sport + 1 - 1024) % (32768 - 1024) + 1024
        if dport is None:
            dport = self.tcp_dport
            self.tcp_dport = (self.tcp_dport + 1 - 1024) % (32768 - 1024) + 1024
        ack = self.tcp_ack_random.randrange(0, 0x100000000) if flags & TH_ACK else 0
        p = IP(len=data_length + 8, p=socket.IPPROTO_TCP, id=self.next_ip_id(), dst=self.ip_dst, src=self.ip_src, data=
               TCP(sport=sport, dport=dport, seq=self.tcp_seq_random.randrange(0, 0x100000000), ack=ack, flags=flags, win=8192, off_x2=(5 << 4), data=
               _fill_data(self, data_length)))
        return p
    def next_packet_testing_mtu(self, mtu, *args, **kwargs):
        return super(TCPSequence, self).next_packet_testing_mtu(mtu - 20, *args, **kwargs)

class TCP6Sequence(IP6Sequence, TCPSequenceMixin):
    """Generates sequences of TCP/IPv6 packets."""
    def __init__(self, ip_dst, ip_src=None, ip_id=None, tcp_sport=None, tcp_dport=None, tcp_seq_random=None, tcp_ack_random=None, data=1):
        super(TCP6Sequence, self).__init__(ip_dst, ip_src, ip_id)
        if tcp_sport is None:
            tcp_sport = int(time.time())
        self.tcp_sport = (tcp_sport - 1024) % (32768 - 1024) + 1024
        if tcp_dport is None:
            tcp_dport = int(time.time())
        self.tcp_dport = (tcp_dport - 1024) % (32768 - 1024) + 1024
        if tcp_seq_random is None:
            tcp_seq_random = random.Random()
        self.tcp_seq_random = tcp_seq_random
        if tcp_ack_random is None:
            tcp_ack_random = random.Random()
        self.tcp_ack_random = tcp_ack_random
        self.data = data & 0xffff
    def next_packet(self, data_length=0, sport=None, dport=None, flags=TCPFlags(TH_SYN)):
        data_length = max(data_length, 0)
        if sport is None:
            sport = self.tcp_sport
            self.tcp_sport = (self.tcp_sport + 1 - 1024) % (32768 - 1024) + 1024
        if dport is None:
            dport = self.tcp_dport
            self.tcp_dport = (self.tcp_dport + 1 - 1024) % (32768 - 1024) + 1024
        ack = self.tcp_ack_random.randrange(0, 0x100000000) if flags & TH_ACK else 0
        p = IP6(plen=data_length + 28, nxt=44, dst=self.ip_dst, src=self.ip_src, p=socket.IPPROTO_TCP, data=
                TCP(sport=sport, dport=dport, seq=self.tcp_seq_random.randrange(0, 0x100000000), ack=ack, flags=flags, win=8192, off_x2=(5 << 4), data=
                _fill_data(self, data_length)))
        p.extension_hdrs[44] = IP6FragmentHeader(nxt=socket.IPPROTO_TCP, id=self.next_ip_id())
        return p
    def next_packet_testing_mtu(self, mtu, *args, **kwargs):
        return super(TCP6Sequence, self).next_packet_testing_mtu(mtu - 20, *args, **kwargs)
