#!/usr/bin/env python

"""Receive TCP/IPv4 replies answering TCP/IPv4 packets."""

__author__ = 'Jeffrey Knockel'
__copyright__ = 'Copyright 2013 Jeffrey Knockel'
__license__ = 'GPL v2'

import sys

from dpkt.dpkt import Packet
from dpkt.tcp import TCP, TH_RST, TH_ACK

from nslookup4 import lookup_or_die
from sequence import TCPSequence
from sniffing import start_sniffing

def receive_tcps(sniffer, sents, timeout):
    """Return a generator yielding received TCP packets answering any packets
    in 'sents' until 'timeout' has expired.

    A TCP packet y answers another x if x.dport == y.sport and x.sport ==
    y.dport and either y ACK's x (y has the ACK flag set and has an ack number
    one greater than x's sequence number) or y has the RST flag set with no ACK
    flag set.
    """
    if isinstance(sents, Packet):
        sents = [sents]
    sents_dict = {}
    for sent in sents:
        sent_tcp = sent.data
        key = sent.dst, sent_tcp.sport, sent_tcp.dport
        sents_dict.setdefault(key, set()).add(sent_tcp.seq + 1 & 0xffffffff)
    while True:
        p = sniffer.sniff(timeout)
        if p is None:
            break
        if type(p.data) is TCP:
            p_tcp = p.data
            key = p.src, p_tcp.dport, p_tcp.sport
            acks = sents_dict.get(key, None)
            if acks is not None:
                if p_tcp.flags & TH_RST and not p_tcp.flags & TH_ACK:
                    yield p
                elif p_tcp.flags & TH_ACK and p_tcp.ack in acks:
                    yield p

def receive_tcp(*args, **kwargs):
    """Return the first TCP packet answering 'sents', or None if 'timeout'
    has expired.

    See receive_tcps for more details.
    """
    return next(receive_tcps(*args, **kwargs), None)

def ping(sniffer, addr, port, mtu):
    seq = TCPSequence(addr, iface=sniffer.iface)
    timeout = sniffer.timeout()
    while True:
        timeout.add_time(1.0)
        request = seq.next_packet_testing_mtu(mtu, dport=port)
        sniffer.send(request, True)
        p = None
        for p in receive_tcps(sniffer, request, timeout):
            print '%s: %d' % (p.data.flags, p.id)
        if p is None:
            print '*'
        sys.stdout.flush()

def usage(code=1):
    sys.stderr.write('Usage: %s HOST PORT MTU\n' % sys.argv[0])
    exit(code)

def main():
    try:
        if len(sys.argv) != 4:
            usage()
        host = sys.argv[1]
        try:
            port = int(sys.argv[2])
            mtu = int(sys.argv[3])
        except ValueError:
            usage()
        addr = lookup_or_die(host)[0]
        start_sniffing(ping, False, args=(addr, port, mtu))
    except KeyboardInterrupt:
        sys.stderr.write('Received SIGINT\n')

if __name__ == '__main__':
    main()
